#include<RealPCube.h>


//====================================================================
RealPCube::RealPCube(const int& DevNo)
{

  uint i=0;


  m_uDevNo = DevNo;

  try
    {
      m_uvMod = new int [NL+1];
      m_vTh = new float [NL+1];

      m_vThMax = new float [NL+1];
      m_vThMin = new float [NL+1];

      m_vThdotMax = new float [NL+1];

      m_lvState = new ulong [NL+1];

      m_vhomevel = new float [NL+1];

      m_vRampvel = new float [NL+1];
      m_vRampacc = new float [NL+1];

    }
  catch(std::bad_alloc a)
    {

      std::cerr<<"\033[30;31m"<<"File :: \t"<<__FILE__<<std::endl;
      std::cerr<<"Function :: \t"<<__FUNCTION__<<std::endl;
      std::cerr<<"Error:Buffers of RealPCube could not be initailzed"<<"\033[30m"<<std::endl;
      exit(EXIT_FAILURE);

    }

  //Initializing the module numbers
  for(i=0;i<=NL;i++)
    m_uvMod[i] = i+1;


  // Activate Robot
  char pInitString[] = "ESD:0,250";
  CHECK_ERROR(PCube_openDevice( &m_uDevNo, pInitString ));
 
  //Reset all the modules
  for(i = 0; i <=NL; i++)
    {
		if(i!=5) //kluge skipping module 6
      CHECK_ERROR(PCube_resetModule(m_uDevNo,m_uvMod[i]));
    }


  std::cout<<"\nmodID\t Max_Position\t Min_Position \t Max_Velocity \t  Current_Position\t\n\n";
  std::cout.setf(std::ios::fixed);
  for(i = 0; i <= NL; i++)
    {
		if(i!=5){ //kluge skipping module 6
      CHECK_ERROR(PCube_getMaxPos(m_uDevNo, m_uvMod[i], &m_vThMax[i]));
      CHECK_ERROR(PCube_getMinPos(m_uDevNo, m_uvMod[i], &m_vThMin[i]));
      CHECK_ERROR(PCube_getMaxVel(m_uDevNo, m_uvMod[i], &m_vThdotMax[i]));
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));

      std::cout<< std::setprecision(4)<<std::setw(10)<<m_uvMod[i]<<"\t";
      std::cout<< std::setprecision(4)<<std::setw(10)<< m_vThMax[i] <<"\t";
      std::cout<< std::setprecision(4)<<std::setw(10)<< m_vThMin[i]<<"\t";
      std::cout<< std::setprecision(4)<<std::setw(10)<< m_vThdotMax[i]<<"\t";
      std::cout<< std::setprecision(4)<<std::setw(10)<< m_vTh[i]<<std::endl;
		}
    }

  std::cout.unsetf(std::ios::fixed);


  //Value is selected based on observation (reasonably fast and smooth motion)
  //homing velocity rad/s
  //negative values of joint 1 & 2 are unacceptable, but positive values are taking it away from the zero position.
  m_vhomevel[0] = -0.5f;
  m_vhomevel[1] = -0.2f;
  m_vhomevel[2] = 0.5f;
  m_vhomevel[3] = 0.2f;
  m_vhomevel[4] = 0.5f;
  m_vhomevel[5] = 0.3f;
  m_vhomevel[6] = 0.5f; 



  //values chosen based on observation (reasonably fast and smooth motion)	
  //Velocity for positioning (rad/s)
  m_vRampvel[0] = 0.7f;
  m_vRampvel[1] = 0.5f;
  m_vRampvel[2] = 0.7f;
  m_vRampvel[3] = 0.5f;
  m_vRampvel[4] = 0.7f;
  m_vRampvel[5] = 0.7f;
  m_vRampvel[6] = 1.4f; 

  //Acceleration for positioning (rad/s^2)
  m_vRampacc[0] = 1.5f;
  m_vRampacc[1] = 1.5f;
  m_vRampacc[2] = 1.5f;
  m_vRampacc[3] = 1.5f;
  m_vRampacc[4] = 1.5f;
  m_vRampacc[5] = 1.5f;
  m_vRampacc[6] = 3.0f; 


  newline(2);
  std::cout<<"Robot initialization over ... ready for further tasks ...";
  newline(3);


}
//====================================================================
RealPCube::~RealPCube()
{

  delete []m_uvMod;
  delete []m_vTh;
  delete []m_vThMax;
  delete []m_vThMin;
  delete []m_vThdotMax;

  delete []m_vhomevel;
  delete []m_vRampvel;
  delete []m_vRampacc;

  del(m_lvState);

}
//====================================================================
void RealPCube::CHECK_ERROR(int ret)const
{
  if(ret != 0)
    {

      std::cout<<"\033[30;31m"<<"Error Flag =   "<<ret<<std::endl;

      PCube_resetAll(m_uDevNo) ;
      PCube_haltAll(m_uDevNo) ;
      PCube_closeDevice(m_uDevNo);

      std::cout<<"Exiting the program "<<"\033[30m"<<std::endl;
      exit(EXIT_FAILURE);
    }

  return;
}
//====================================================================
void RealPCube::move_home()
{


  uint i;


  //===========================
  //Reset all the modules
  for(i = 0; i <NL; i++)
    { 
		if(i!=5) //kluge skipping module 6
      CHECK_ERROR(PCube_resetModule(m_uDevNo,m_uvMod[i]));
    }

  std::cout<<"Robot moving to its Home Position ... Watch out!\n\n";


  //===========================
  //Giving Homing command to all the modules 

  for(i = 0; i < NL; i++)
    {
		if(i!=5) //kluge skipping module 6
      CHECK_ERROR(PCube_homeModule(m_uDevNo, m_uvMod[i]));

      usleep(2000);
		
    }


  Sleep(1);

  //===========================
  //Checking whether modules are homed
  for(i = 0; i < NL; i++)
    {

      do
	{
		if(i!=5) //kluge skipping module 6
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &m_lvState[i]));
	  usleep(5000);

	}while( !(m_lvState[i] & STATEID_MOD_HOME) ); 
		if(i!=5) //kluge skipping module 6
      std::cout<<"Module    "<<m_uvMod[i]<<" is homed .. "<<std::endl;
      usleep(2000);
		
    }


  std::cout<<"\n\nRobot is now at Home Position\n\n";

}
//====================================================================
void RealPCube::check_move_home()
{


  uint i=0;


  //=============================================
  //Checking whether robot is homed already , if not so homing command is issued
  printf("Robot Checking the Home status and homing command will issues for unhomed modules ... \n\n");


  for(i = 0; i < NL; i++)
    {
		if(i!=5) {//kluge skipping module 6
      CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &m_lvState[i]));


      if( !(m_lvState[i] & STATEID_MOD_HOME) )
	{
	  std::cout<<"Module "<<m_uvMod[i]<< "  is not homed..."<<std::endl;
	  CHECK_ERROR(PCube_resetModule(m_uDevNo,m_uvMod[i]));
	  usleep(2000);
	  CHECK_ERROR(PCube_setHomeVel(m_uDevNo,m_uvMod[i],m_vhomevel[i]));
	  usleep(2000);
	  CHECK_ERROR(PCube_homeModule(m_uDevNo, m_uvMod[i]));
	}

      usleep(2000);
		}
    }

  Sleep(1);


  //===========================================
  //Checking whether modules are homed after checking the modules come to rest
  for(i = 0; i < NL; i++)
    {

		if(i!=5) {//kluge skipping module 6
      //checking whether the Module is at rest
      do
	{

	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &m_lvState[i]));

	  usleep(5000);
	}while(!(m_lvState[i] & STATEID_MOD_RAMP_END ));


      //checking for home position
      usleep(10000);
      CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &m_lvState[i])); 

      if( !(m_lvState[i] & STATEID_MOD_HOME) )
	{
	  std::cout<<"\033[30;32m"<<"Module"<< m_uvMod[i]<<" is not homed ... "<<std::endl;
	  ManipulatorStop();
	  std::cout<<"Manipulator is halted"<<"\033[30m"<<std::endl;
	  exit(-1);
	}
      else
	std::cout<<"Module  "<< m_uvMod[i]<<" is homed ... "<<std::endl;

		}
    }
  newline(2);

  
}
//====================================================================
void RealPCube::Move_module2home(const int& modID)const
{
  ulong state = 0;

  CHECK_ERROR(PCube_homeModule(m_uDevNo, modID));

  do
    { 
      CHECK_ERROR(PCube_getModuleState(m_uDevNo, modID, &state));
      usleep(1000);
    }while(!(state & STATEID_MOD_HOME));

  std::cout<<"Module  "<<modID<<"  is now at Home Position"<<std::endl;
}
//====================================================================
void RealPCube::ManipulatorStop()const
{
  uint i;
  float Thetadot[7]; 

  //============================
  //Initializing low velocity
  for(i =0;i<7;i++)
    {
      Thetadot[i] = 0.0005;
    }

  //===============================
  //Halting the device
  MoveVelocity(Thetadot);
  CHECK_ERROR(PCube_haltAll(m_uDevNo) );
  CHECK_ERROR(PCube_resetAll(m_uDevNo) );

  std::cout<<"PowerCube has been stopped successfully\n"<<std::endl;


  return;
}
//====================================================================
void RealPCube::Robot_close()const
{
  CHECK_ERROR(PCube_closeDevice(m_uDevNo));
 
  std::cout<<"PowerCube Robot Device has been closed successfully"<<std::endl;

}

//====================================================================
void RealPCube::Check_MovePosition(const float* const vTh)
{


  uint i=0;
  ulong state=0;

	

  //===========================
  // Move to the Position
  for(i =  0; i < NL; i++)
    {
		if(i!=5) {//kluge skipping module 6
      CHECK_ERROR(PCube_setRampVel(m_uDevNo, m_uvMod[i], m_vRampvel[i]));
      usleep(5000);
      CHECK_ERROR(PCube_setRampAcc(m_uDevNo, m_uvMod[i], m_vRampacc[i]));
      usleep(5000);
      CHECK_ERROR(PCube_movePos(m_uDevNo, m_uvMod[i], vTh[i]));

      usleep(10000);
		}
    }


  Sleep(2);

  //==========================
  //Checking whether modules are moved and getting the new joint angle
  for (i=0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      do
	{
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &state));
	  usleep(5000);

	}while( state & STATEID_MOD_MOTION);
		

    }
  Sleep(1);
  //================================

  for(i=0;i<NL;i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));
      usleep(5000);

      if( fabs(vTh[i]) > 0.02)
	{
	  if( (fabs(m_vTh[i]) < 0.98*fabs(vTh[i]))  || (fabs(m_vTh[i]) > 1.02*fabs(vTh[i])) )
	    {
	      // std::cout<<"Module  " << m_uvMod[i] << "  is not reached. Trying again \n";
	      CHECK_ERROR(PCube_setRampVel(m_uDevNo, m_uvMod[i], m_vRampvel[i]));
	      usleep(5000);
	      CHECK_ERROR(PCube_setRampAcc(m_uDevNo, m_uvMod[i], m_vRampacc[i]));
	      usleep(5000);
	      CHECK_ERROR(PCube_movePos(m_uDevNo, m_uvMod[i], vTh[i]));
	      usleep(10000);

	    }

	}
      else
	{

	  if( (fabs(m_vTh[i]) > 0.02) )
	    {
	      // std::cout<<"Module  " << m_uvMod[i] << "  is not reached. Trying again \n";
	      CHECK_ERROR(PCube_setRampVel(m_uDevNo, m_uvMod[i], m_vRampvel[i]));
	      usleep(5000);
	      CHECK_ERROR(PCube_setRampAcc(m_uDevNo, m_uvMod[i], m_vRampacc[i]));
	      usleep(5000);
	      CHECK_ERROR(PCube_movePos(m_uDevNo, m_uvMod[i], vTh[i]));
	      usleep(10000);

	    }
	}
    }
  Sleep(2);

  //==========================
  //Getting the new joint angle
  for (i=0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      do
	{
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &state));
	  usleep(5000);

	}while( state & STATEID_MOD_MOTION);
		
      usleep(10000);
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));


      if( fabs(vTh[i]) > 0.02)
	{
	  if( (fabs(m_vTh[i]) < 0.98*fabs(vTh[i]))  || (fabs(m_vTh[i]) > 1.02*fabs(vTh[i])) )
	    {

	      std::cerr<<"\033[30;32m"<<"Module  "<<m_uvMod[i]<<"  has not reached the desired position "<<"\033[30m"<<std::endl;

	    }

	}
      else
	{

	  if( (fabs(m_vTh[i]) > 0.02) )
	    {


	      std::cerr<<"\033[30;32m"<<"Module  "<<m_uvMod[i]<<"  has not reached the desired position:: "<<"\033[30m"<<std::endl;


	    }
	}		
    }

  return;

}
//====================================================================
void RealPCube::MovePosition(const float* const vTh)
{


  int i=0;
  ulong state=0;
	

  //===========================
  // Move to the Position
  for(i =  0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_setRampVel(m_uDevNo, m_uvMod[i], m_vRampvel[i]));
      CHECK_ERROR(PCube_setRampAcc(m_uDevNo, m_uvMod[i], m_vRampacc[i]));
      CHECK_ERROR(PCube_movePos(m_uDevNo, m_uvMod[i], vTh[i]));

      usleep(5000);
    }


  Sleep(2);
  //==========================
  //Getting the new joint angle
  for (i=0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      do
	{
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &state));
	  usleep(2000);

	}while( state & STATEID_MOD_MOTION);
		
      usleep(5000);
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));
		
    }

  return;

}
//====================================================================
void RealPCube::Move_module_pos(const int& mod, const float& Th)const 
{

  ulong state=0;

	
  CHECK_ERROR(PCube_setRampVel(m_uDevNo, mod, 0.7));
  CHECK_ERROR(PCube_setRampAcc(m_uDevNo, mod, 0.5));
	
  CHECK_ERROR(PCube_movePos(m_uDevNo, mod, Th));
  do
    {
      CHECK_ERROR(PCube_getModuleState(m_uDevNo, mod, &state));
      usleep(2000);

    }while( (state & STATEID_MOD_MOTION) );
	
  CHECK_ERROR(PCube_getPos(m_uDevNo, mod, &m_vTh[mod-1]));
	

}
	
//====================================================================
void RealPCube::MoveVelocity(const float* const Thdot)const
{
  uint i=0;
	

  // Move to the Position
  for(i = 0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      //CHECK_ERROR(PCube_resetModule(dev,mod[i]));
      CHECK_ERROR(PCube_moveVel(m_uDevNo, m_uvMod[i], Thdot[i]));
    }

}
//====================================================================
void RealPCube::read_joint_angles(float* const vTh)
{
  uint i=0;


	
  for(i = 0; i < 7; i++)
    {

		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));
      vTh[i] = m_vTh[i];
    }

}


//====================================================================
void RealPCube::read_joint_velocity(float* const vThdot)const
{
  uint i=0;


	
  for(i = 0; i < 7; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_getVel(m_uDevNo, m_uvMod[i], &vThdot[i]));
    }

}


//====================================================================
void RealPCube::check_angle(const float* const vTh, float* const vThLimit)const
{

  for(uint i =0; i < NL;i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      if(vTh[i] > m_vThMax[i])
	{
	  std::cout<<"\033[30;32m"<< "Joint  "<<(i+1)<<"  exceeds the maximum limit"<<"\033[30m"<<std::endl;
	  vThLimit[i] = m_vThMax[i];
	}
      else if(vTh[i] < m_vThMin[i])
	{
	  std::cout<<"\033[30;32m"<< "Joint  "<<(i+1)<<"  exceeds the minimum limit"<<"\033[30m"<<std::endl;
	  vThLimit[i] = m_vThMin[i];
	}
      else
	vThLimit[i] = vTh[i];


    }

  return;
}



//====================================================================
void RealPCube::check_velocity(const float* const vThdot, float* const vThdotLimit)const
{


  for(uint i =0; i < NL;i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      if(vThdot[i] > m_vThdotMax[i])
	{
	  std::cout<<"\033[30;32m"<< "Joint  "<<(i+1)<<"  exceeds the maximum limit"<<"\033[30m"<<std::endl;
	  vThdotLimit[i] = m_vThdotMax[i];
	}
      else if(vThdot[i] < -m_vThdotMax[i])
	{
	  std::cout<<"\033[30;32m"<< "Joint  "<<(i+1)<<"  exceeds the minimum limit"<<"\033[30m"<<std::endl;
	  vThdotLimit[i] = -m_vThdotMax[i];
	}
      else
	vThdotLimit[i] = vThdot[i];


    }


  return;
}
//====================================================================
void RealPCube::getJointAngleLimit(float* const vThMax, float* const vThMin)
{

  copy(m_vThMax,vThMax,NL);
  copy(m_vThMin,vThMin,NL);

  return;
}
//====================================================================
void RealPCube::reset()
{

  int i;
  //Reset all the modules
  for(i = 0; i <=NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_resetModule(m_uDevNo,m_uvMod[i]));
    }


  return;

}
//====================================================================
void RealPCube::moveStep(const float* const vTh, const long int& sTime)
{


  for(int i =  0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_moveStep(m_uDevNo, m_uvMod[i], vTh[i], sTime)) ;
      usleep(5000);
    }

  return ;
}
//====================================================================
void RealPCube::check_moveStep(const float* const vTh, const long int& sTime)
{


  uint i=0;
  ulong state=0;
	

  //===========================
  // Move to the Position

  moveStep(vTh, sTime);

  usleep(sTime* 1000);

  //==========================
  //Checking whether modules are moved and getting the new joint angle
  for (i=0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      do
	{
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &state));
	  usleep(5000);

	}while( state & STATEID_MOD_MOTION);
		

    }
  //================================

  for(i=0;i<NL;i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));
      usleep(5000);

      if( fabs(vTh[i]) > 0.02)
	{
	  if( (fabs(m_vTh[i]) < 0.98*fabs(vTh[i]))  || (fabs(m_vTh[i]) > 1.02*fabs(vTh[i])) )
	    {
	      CHECK_ERROR(PCube_moveStep(m_uDevNo, m_uvMod[i], vTh[i], sTime) );
	      usleep(5000);

	    }

	}
      else
	{

	  if( (fabs(m_vTh[i]) > 0.02) )
	    {
	      CHECK_ERROR(PCube_moveStep(m_uDevNo, m_uvMod[i], vTh[i], sTime) );
	      usleep(5000);

	    }
	}
    }
  usleep(sTime*1000);

  //==========================
  //Checking whether modules are moved and getting the new joint angle

  for (i=0; i < NL; i++)
    {
		//if(i==5)i++; //kluge skipping module 6
      do
	{
	  CHECK_ERROR(PCube_getModuleState(m_uDevNo, m_uvMod[i], &state));
	  usleep(5000);

	}while( state & STATEID_MOD_MOTION);
		
      usleep(5000);

      CHECK_ERROR(PCube_getPos(m_uDevNo, m_uvMod[i], &m_vTh[i]));


      if( fabs(vTh[i]) > 0.02)
	{
	  if( (fabs(m_vTh[i]) < 0.98*fabs(vTh[i]))  || (fabs(m_vTh[i]) > 1.02*fabs(vTh[i])) )
	    {

	      std::cerr<<"\033[30;32m"<<"Module  "<<m_uvMod[i]<<"  has not reached the desired position "<<"\033[30m"<<std::endl;

	    }

	}
      else
	{

	  if( (fabs(m_vTh[i]) > 0.02) )
	    {


	      std::cerr<<"\033[30;32m"<<"Module  "<<m_uvMod[i]<<"  has not reached the desired position:: "<<"\033[30m"<<std::endl;


	    }
	}		
    }

  return;

}
//====================================================================
//EOF


 
  