/*
//====================================================================
File to perform matrix operations using gsl
16.01.2009
 i) Psuedo Inverse computation routine is included

//====================================================================

 */

#ifndef _OPERATIONS__
#define _OPERATIONS_


#include<constants.h>

#include<gsl/gsl_matrix.h>
#include<gsl/gsl_eigen.h>
#include<gsl/gsl_linalg.h>
#include<gsl/gsl_blas.h>
#include<gsl/gsl_permutation.h>

//====================================================================
//earlier right inverse routine
void PInverse(const float* const J, const uint& row, const uint& col,  float* const JI);


//computation of inverse with singular value decomposition
void inverseSV(const float* A, const uint& order, float* const AI);


//computation of inverse with LU decomposition
void inverseLU(const float *A, const uint& order, float* const AI);

//computation of right inverse inv(A) = A(AA^T){-1}  (shouter/shorter matrix)
void RInverse(const float* const A, const uint& row, const uint& col, float* const AI);


//computation of left inverse inv(A) = (A^TA)^{-1}A^T (lengthier matrix)
void LInverse(const float* const A, const uint& row, const uint& col, float* const AI);


//copies the matrix to gsl_matrix
void copy(const float* const& A, gsl_matrix* const gA, const uint& row, const uint& col, const uint& ATrans =  NO_TRANS);

//copies buffer from gsl_matrix to matrix
void copy(const gsl_matrix* const& gA, float* const& A, const uint& row, const uint& col, const uint& ATrans = NO_TRANS);


//copies the buffer from gsl_vector to vector
void copy(const gsl_vector* const& gA, float* const& A, const uint& col);


//copies the buffer from vector to gsl_vector
void copy(const float* const& A, gsl_vector* const gA, const uint& col);


//display a gsl matrix
void display(const gsl_matrix* const& gA, const uint& row, const uint& col);



//display a gsl vector
void display(const gsl_vector* const& gA, const uint& col);

//compute eigen value and vector of matrix
void computeEigen(const float* const& A, const uint& order, float* const& eigenval=NULL, float* const& eigvec=NULL);


#endif
