/* ----------------------------------------------------------
 * Header file for PowerCube Manipulator(with orientation)
 * available at IIT-Kanpur
 * Date: December 2009
 * ---------------------------------------------------------- */
#include<PCube.h>



//====================================================================
PCube::PCube()
{

  initDimensionConstants();

  // Joint Angles
  q1=0;
  q2=0;
  q3=0;
  q4=0;
  q5=0;
  q6=0;
  q7=0;

  createBuffer();
  m_vplanePt[NC] = 1;


  //Camera Rotation Matrix
  RotationMatrix(alpha8,theta8,m_mR8);

  //Camera Transformation Matrix
  TransformationMatrix(a8,alpha8,d8,theta8,m_mT8);

  std::cout<<"PowerCube Kinematics Initialized\n";

#ifdef DEBUG
  PCubeDisplay();
#endif



}
//=====================================================================
void PCube::initDimensionConstants()
{

  //link length
  d1 = 0.368f;
  d3 = 0.3815f;
  d5 = 0.3085f;
  d7 = 0.2656f; 


  //Camera Kinematic Parameters
  a8 = A8;
  alpha8 = ALPHA8;
  d8 = D8;
  theta8 = THETA8;


  return;
}
//=====================================================================
void PCube::createBuffer()
{

  try
    {

      m_vplaneNormal = new float [NC];
      m_vplanePt = new float [NC+1];

      m_mR8 = new float [NO*NO];
      m_mT8 = new float [(NC+1) * (NC+1)];


      m_mR = new float [NO*NO];
      m_mT = new float [(NC+1) * (NC+1)];

      m_mJp = new float [NC * NL];
      m_mJo = new float [NO * NL];


      m_mJpe = new float [NC * NL];
      m_mJoe = new float [NO * NL];


    }
  catch(std::bad_alloc a)
    {

      std::cerr<<"\033[30;31m"<<"File :: \t"<<__FILE__<<std::endl;
      std::cerr<<"Function :: \t"<<__FUNCTION__<<std::endl;
      std::cerr<<"Error:Buffers of PCube could not be initailzed"<<"\033[30m"<<std::endl;
      exit(EXIT_FAILURE);

    }

  return;

}
//=====================================================================
void PCube::PCubeDisplay()
{

  std::cout<<"Camera Rotation Matrix:: \n";
  display(m_mR8,NO,NO);

  std::cout<<"Camera Transformation Matrix:: \n";
  display(m_mT8,NO+1,NO+1);


}
//=====================================================================
PCube::~PCube()
{


  delete []m_mR8;
  delete []m_mT8;

  delete []m_mR;
  delete []m_mT;


  delete []m_mJp;
  delete []m_mJo;

  delete []m_mJpe;
  delete []m_mJoe;

  delete []m_vplaneNormal;
  delete []m_vplanePt;



  std::cout<<"PowerCube Kinematics is cleared";
  newline();

}
//=====================================================================
void PCube::initAngle(const float th[])
{

  //=================================
  // Variable Initialization
  q1 = th[0];
  q2 = th[1];
  q3 = th[2];
  q4 = th[3];
  q5 = th[4];
  q6 = th[5];
  q7 = th[6];

  return;

}
//=====================================================================
void PCube::PositionKinematics(const float* const vTh, float* const pos)
{

  // Forward Position Kinematics
  // Gives Cartesian Position of end-effector ( x, y, z)
  // Input : th[7] (joint Angles)
  // Output : pos[3] (Cartesian Postion of end-effector)


  float x;                 // Cartesian position (x)             
  float y;                 // Cartesian position (y)
  float z;                 // Cartesian position (z)


  //=================================
  // Variable Initialization

  initAngle(vTh);


  //=================================
  // forward Kinematics (x in metre)
  x = -d7*((-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	    -cos(q1)*sin(q2)*cos(q4))
           *cos(q6)
           -((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q5)
	     +((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
	       -cos(q1)*sin(q2)*sin(q4))
             *cos(q5))
	   *sin(q6))
    +d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	 +cos(q1)*sin(q2)*cos(q4))+d3*cos(q1)*sin(q2);



  //=================================
  // forward Kinematics (yin metre)
  y = -d7*((-(cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	    -sin(q1)*sin(q2)*cos(q4))
           *cos(q6)
           -((cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q5)
	     +((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
	       -sin(q1)*sin(q2)*sin(q4))
             *cos(q5))
	   *sin(q6))
    +d5*((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	 +sin(q1)*sin(q2)*cos(q4))+d3*sin(q1)*sin(q2);

  //=================================
  // forward Kinematics (z in metre)

  z = -d7*((sin(q2)*cos(q3)*sin(q4)-cos(q2)*cos(q4))*cos(q6)
           -(sin(q2)*sin(q3)*sin(q5)+(-cos(q2)*sin(q4)
				      -sin(q2)*cos(q3)*cos(q4))
	     *cos(q5))
	   *sin(q6))
    +d5*(cos(q2)*cos(q4)-sin(q2)*cos(q3)*sin(q4))+d3*cos(q2)+d1;

  //===================================
  // Storing in destination Buffer
  pos[0] = x;
  pos[1] = y;
  pos[2] = z;


  return;

}

//=====================================================================

void PCube::RotationMatrix(const float* const vTh)
{



  initAngle(vTh);

  //======================================================
  //R(1,1)
  m_mR[0] = cos(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)
				    *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
				    +sin(q4)*sin(q6)*sin(q7))
				  -sin(q3)
				  *(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7)))
			 +sin(q2)*(cos(q4)*sin(q6)*sin(q7)
				   -sin(q4)
				   *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))))
    -sin(q1)*(sin(q3)*(cos(q4)
		       *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
		       +sin(q4)*sin(q6)*sin(q7))
	      +cos(q3)*(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7)));



  //R(1,2)
  m_mR[1] = cos(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)
				    *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
				    -sin(q4)*sin(q6)*cos(q7))
				  -sin(q3)
				  *(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7)))
			 +sin(q2)*(-sin(q4)
				   *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
				   -cos(q4)*sin(q6)*cos(q7)))
    -sin(q1)*(sin(q3)*(cos(q4)
		       *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		       -sin(q4)*sin(q6)*cos(q7))
	      +cos(q3)*(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7)));



  //R(1,3)
  m_mR[2] = cos(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
				  -sin(q3)*sin(q5)*sin(q6))
			 +sin(q2)*(cos(q4)*cos(q6)-sin(q4)*cos(q5)*sin(q6)))
    -sin(q1)*(sin(q3)*(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
	      +cos(q3)*sin(q5)*sin(q6));





  //======================================================

  //R(2,1)
  m_mR[3] = sin(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)
				    *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
				    +sin(q4)*sin(q6)*sin(q7))
				  -sin(q3)
				  *(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7)))
			 +sin(q2)*(cos(q4)*sin(q6)*sin(q7)
				   -sin(q4)
				   *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))))
    +cos(q1)*(sin(q3)*(cos(q4)
		       *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
		       +sin(q4)*sin(q6)*sin(q7))
	      +cos(q3)*(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7)));




  //R(2,2)
  m_mR[4] = sin(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)
				    *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
				    -sin(q4)*sin(q6)*cos(q7))
				  -sin(q3)
				  *(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7)))
			 +sin(q2)*(-sin(q4)
				   *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
				   -cos(q4)*sin(q6)*cos(q7)))
    +cos(q1)*(sin(q3)*(cos(q4)
		       *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		       -sin(q4)*sin(q6)*cos(q7))
	      +cos(q3)*(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7)));




  //R(2,3)
  m_mR[5] = sin(q1)*(cos(q2)*(cos(q3)
				  *(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
				  -sin(q3)*sin(q5)*sin(q6))
			 +sin(q2)*(cos(q4)*cos(q6)-sin(q4)*cos(q5)*sin(q6)))
    +cos(q1)*(sin(q3)*(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
	      +cos(q3)*sin(q5)*sin(q6));


  //=====================================================


  //R(3,1)
  m_mR[6] = cos(q2)*(cos(q4)*sin(q6)*sin(q7)
			 -sin(q4)*(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7)))
    -sin(q2)*(cos(q3)*(cos(q4)
		       *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
		       +sin(q4)*sin(q6)*sin(q7))
	      -sin(q3)*(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7)));



  //R(3,2)
  m_mR[7] = cos(q2)*(-sin(q4)*(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
			 -cos(q4)*sin(q6)*cos(q7))
    -sin(q2)*(cos(q3)*(cos(q4)
		       *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		       -sin(q4)*sin(q6)*cos(q7))
	      -sin(q3)*(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7)));





  //R(3,3)
  m_mR[8] = cos(q2)*(cos(q4)*cos(q6)-sin(q4)*cos(q5)*sin(q6))
    -sin(q2)*(cos(q3)*(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
	      -sin(q3)*sin(q5)*sin(q6));


 
  //==============================================================

  return;

}

//=========================================================================
void PCube::TransformationMatrix(const float* const vTh)
{

  float x[NC];


  RotationMatrix(vTh);

  PositionKinematics(vTh,x);

  //Rotation Matrix
  for(uint i=0; i < NO; i++ )
    for(uint j=0; j < NO; j++)
      m_mT[i*(NO+1)+j] = m_mR[i*(NO)+j];

  //position vector
  for(uint i=0; i < NC; i++)
    m_mT[i*(NO+1)+3] = x[i];

  //Projection vector
  for(uint i=0;i < NO;i++)
    m_mT[3*(NO+1) + i] = 0;

  //Position constant
  m_mT[3*(NO+1) + 3] = 1;

  return;

}
//=========================================================================
void PCube::RotationMatrix(const float* const vTh, float* const mR)
{

  RotationMatrix(vTh);
  copy(m_mR,mR,NO,NO);

  return;
}
//=========================================================================
void PCube::TransformationMatrix(const float* const vTh, float* const mT)
{

  TransformationMatrix(vTh);
  copy(m_mT,mT,(NO+1),(NO+1));

  return;
}
//=========================================================================
void PCube::TransformationMatrix(const float& length, const float& twist, const float& offset, const float& angle, float * const T)const
{

  //  T(1,1)
  T[0] = cos(angle);
  //  T(1,2)
  T[1] = -sin(angle) * cos(twist); 
  //  T(1,3)
  T[2] = sin(angle) * sin(twist);
  //  T(1,4)
  T[3] = length*cos(angle);


  // T(2,1)
  T[4] = sin(angle);
  // T(2,2)
  T[5] = cos(angle)*cos(twist);
  // T(2,3)
  T[6] = -cos(angle)*sin(twist);
  // T(2,4)
  T[7] = length*sin(angle);


  // T(3,1)
  T[8] = 0;
  // T(3,2)
  T[9] = sin(twist);
  // T(3,3)
  T[10] = cos(twist);
  // T(3,4)
  T[11] = offset;


  // T(4,1)
  T[12] = 0;
  // T(4,2)
  T[13] = 0;
  // T(4,3)
  T[14] = 0;
  // T(4,4)
  T[15] = 1;

}
//=====================================================================
void PCube::RotationMatrix(const float& twist, const float& angle, float * const R)const
{


  //  R(1,1)
  R[0] = cos(angle);
  //  R(1,2)
  R[1] = -sin(angle) * cos(twist);
  //  R(1,3)
  R[2] = sin(angle) * sin(twist);


  //===============================

  //  R(2,1)
  R[3] = sin(angle);
  //  R(2,2)
  R[4] = cos(angle)*cos(twist);
  //  R(2,3)
  R[5] = -cos(angle)*sin(twist);


  //===============================

  //  R(3,1)
  R[6] = 0;
  //  R(3,2)
  R[7] = sin(twist);
  //  R(2,3)
  R[8] = cos(twist);


  return;
}
//=====================================================================
void PCube::PositionJacobian(const float* const vTh, float* const J)
{



  initAngle(vTh);

  //====================== Jacobian -x =============================
  //J(1,1)
  J[0] = -d7*((sin(q1)*sin(q2)*cos(q4)-(-cos(q1)*sin(q3)
					   -sin(q1)*cos(q2)*cos(q3))
		  *sin(q4))
		 *cos(q6)
		 -((sin(q1)*cos(q2)*sin(q3)-cos(q1)*cos(q3))*sin(q5)
		   +(sin(q1)*sin(q2)*sin(q4)+(-cos(q1)*sin(q3)
					      -sin(q1)*cos(q2)*cos(q3))
		     *cos(q4))
		   *cos(q5))
		 *sin(q6))
    +d5*((-cos(q1)*sin(q3)-sin(q1)*cos(q2)*cos(q3))*sin(q4)
	 -sin(q1)*sin(q2)*cos(q4))-d3*sin(q1)*sin(q2);


  //J(1,2)
  J[1] = -d7*((cos(q1)*sin(q2)*cos(q3)*sin(q4)-cos(q1)*cos(q2)*cos(q4))*cos(q6)
		 -(cos(q1)*sin(q2)*sin(q3)*sin(q5)
		   +(-cos(q1)*cos(q2)*sin(q4)-cos(q1)*sin(q2)*cos(q3)*cos(q4))
		   *cos(q5))
		 *sin(q6))
    +d5*(cos(q1)*cos(q2)*cos(q4)-cos(q1)*sin(q2)*cos(q3)*sin(q4))
    +d3*cos(q1)*cos(q2);


  //J(1,3)
  J[2] = d5*(-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q4)
    -d7*(-((sin(q1)*sin(q3)-cos(q1)*cos(q2)*cos(q3))*sin(q5)
	   +(-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*cos(q4)*cos(q5))
	 *sin(q6)
	 -(-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q4)*cos(q6));



  //J(1,4)
  J[3] = d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
		-cos(q1)*sin(q2)*sin(q4))
    -d7*((cos(q1)*sin(q2)*sin(q4)-(cos(q1)*cos(q2)*cos(q3)
				   -sin(q1)*sin(q3))
	  *cos(q4))
	 *cos(q6)
	 -(-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	   -cos(q1)*sin(q2)*cos(q4))
	 *cos(q5)*sin(q6));



  //J(1,5)
  J[4] = d7*((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*cos(q5)
		-((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
		  -cos(q1)*sin(q2)*sin(q4))
		*sin(q5))*sin(q6);


  //J(1,6)
  J[5] = -d7*(-(-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
		   -cos(q1)*sin(q2)*cos(q4))
		 *sin(q6)
		 -((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q5)
		   +((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
		     -cos(q1)*sin(q2)*sin(q4))
		   *cos(q5))
		 *cos(q6));


  //J(1,7)
  J[6] = 0;





  //==================== Jacobian - y ===================================================

  //J(2,1)
  J[7] = -d7*((-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
		  -cos(q1)*sin(q2)*cos(q4))
		 *cos(q6)
		 -((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q5)
		   +((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
		     -cos(q1)*sin(q2)*sin(q4))
		   *cos(q5))
		 *sin(q6))
    +d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	 +cos(q1)*sin(q2)*cos(q4))+d3*cos(q1)*sin(q2);


  //J(2,2)
  J[8] = -d7*((sin(q1)*sin(q2)*cos(q3)*sin(q4)-sin(q1)*cos(q2)*cos(q4))*cos(q6)
		 -(sin(q1)*sin(q2)*sin(q3)*sin(q5)
		   +(-sin(q1)*cos(q2)*sin(q4)-sin(q1)*sin(q2)*cos(q3)*cos(q4))
		   *cos(q5))
		 *sin(q6))
    +d5*(sin(q1)*cos(q2)*cos(q4)-sin(q1)*sin(q2)*cos(q3)*sin(q4))
    +d3*sin(q1)*cos(q2);


  //J(2,3)
  J[9] = d5*(cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q4)
    -d7*(-((-cos(q1)*sin(q3)-sin(q1)*cos(q2)*cos(q3))*sin(q5)
	   +(cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*cos(q4)*cos(q5))
	 *sin(q6)
	 -(cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q4)*cos(q6));



  //J(2,4)
  J[10] = d5*((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
		-sin(q1)*sin(q2)*sin(q4))
    -d7*((sin(q1)*sin(q2)*sin(q4)-(cos(q1)*sin(q3)
				   +sin(q1)*cos(q2)*cos(q3))
	  *cos(q4))
	 *cos(q6)
	 -(-(cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	   -sin(q1)*sin(q2)*cos(q4))
	 *cos(q5)*sin(q6));


  //J(2,5)
  J[11] = d7*((cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*cos(q5)
		-((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
		  -sin(q1)*sin(q2)*sin(q4))
		*sin(q5))*sin(q6);



  //J(2,6)
  J[12] = -d7*(-(-(cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
		   -sin(q1)*sin(q2)*cos(q4))
		 *sin(q6)
		 -((cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q5)
		   +((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
		     -sin(q1)*sin(q2)*sin(q4))
		   *cos(q5))
		 *cos(q6));


  //J(2,7)
  J[13] = 0;


  //======================= Jacobian - z ===========================================

  //J(3,1)
  J[14] = 0;


  //J(3,2)
  J[15] = -d7*((cos(q2)*cos(q3)*sin(q4)+sin(q2)*cos(q4))*cos(q6)
		 -(cos(q2)*sin(q3)*sin(q5)+(sin(q2)*sin(q4)
					    -cos(q2)*cos(q3)*cos(q4))
		   *cos(q5))
		 *sin(q6))
    +d5*(-cos(q2)*cos(q3)*sin(q4)-sin(q2)*cos(q4))-d3*sin(q2);



  //J(3,3)
  J[16] = d5*sin(q2)*sin(q3)*sin(q4)-d7*(-(sin(q2)*cos(q3)*sin(q5)
					     +sin(q2)*sin(q3)*cos(q4)*cos(q5))
					   *sin(q6)
					   -sin(q2)*sin(q3)*sin(q4)*cos(q6));



  //J(3,4)
  J[17] = d5*(-cos(q2)*sin(q4)-sin(q2)*cos(q3)*cos(q4))
    -d7*((cos(q2)*sin(q4)+sin(q2)*cos(q3)*cos(q4))*cos(q6)
	 -(sin(q2)*cos(q3)*sin(q4)-cos(q2)*cos(q4))*cos(q5)*sin(q6));



  //J(3,5)
  J[18] = d7*(sin(q2)*sin(q3)*cos(q5)-(-cos(q2)*sin(q4)-sin(q2)*cos(q3)*cos(q4))
		*sin(q5))*sin(q6);


  //J(3,6)
  J[19] = -d7*(-(sin(q2)*cos(q3)*sin(q4)-cos(q2)*cos(q4))*sin(q6)
		 -(sin(q2)*sin(q3)*sin(q5)+(-cos(q2)*sin(q4)
					    -sin(q2)*cos(q3)*cos(q4))
		   *cos(q5))
		 *cos(q6));


  //J(3,7)
  J[20] = 0;


  //==================================================================


  return;

}
//=================================================================
void PCube::PositionJacobianEndeffector(const float* const vTh, float* const J)
{


  PositionJacobian(vTh,m_mJp);

  RotationMatrix(vTh);

#ifdef DEBUG
  std::cout<<"Rotation Matrix::\n";
  display(m_mR,NO,NO);

#endif

  // RT * J {multiplication is done along row-wise of R matrix}
  multiply(m_mR,m_mJp,J,3,7,3,TRANS);

  return;

}
//=========================================================================
void PCube::PositionJacobianCamera(const float* const vTh, float* const J)
{



  PositionJacobianEndeffector(vTh,m_mJpe);

#ifdef DEBUG
  std::cout<<"Image Rotation Matrix::\n";
  display(m_mR8,NO,NO);
#endif

  // RT * J {multiplication is done along row-wise of R matrix}
  multiply(m_mR8,m_mJpe,J,3,7,3,TRANS);


  return;

}
//=========================================================================
void PCube::OrientationJacobian(const float* const vTh, float* const J)
{


  initAngle(vTh);  

  //==========================================================

  //J(1,1) 
  J[0] = 0;
  //J(2,1)
  J[7] = 0;
  //J(3,1)
  J[14] = 1;

  //===============================

  //J(1,2)
  J[1] = -sin(q1);
  //J(2,2)
  J[8] = cos(q1);
  //J(3,2)
  J[15] = 0;


  //===============================

  //J(1,3)
  J[2] = cos(q1)*sin(q2);
  //J(2,3)
  J[9] = sin(q1)*sin(q2);
  //J(3,3)
  J[16] = cos(q2);


  //===============================

  //J(1,4)
  J[3] = -cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3);

  //J(2,4)
  J[10] = cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3);


  //J(3,4)
  J[17] = sin(q2)*sin(q3);

  //===============================

  //J(1,5)
  J[4] = (cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
    +cos(q1)*sin(q2)*cos(q4);

  //J(2,5)
  J[11] = (cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
    +sin(q1)*sin(q2)*cos(q4);

  //J(3,5)
  J[18] = cos(q2)*cos(q4)-sin(q2)*cos(q3)*sin(q4);


  //===============================

  //J(1,6)
  J[5] = (-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*cos(q5)
    -((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
      -cos(q1)*sin(q2)*sin(q4))
    *sin(q5);

  //J(2,6)
  J[12] = (cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*cos(q5)
    -((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
      -sin(q1)*sin(q2)*sin(q4))
    *sin(q5);


  //J(3,6)
  J[19] = sin(q2)*sin(q3)*cos(q5)-(-cos(q2)*sin(q4)
				     -sin(q2)*cos(q3)*cos(q4))
    *sin(q5);



  //===============================


  //J(1,7)
  J[6] = (-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	     -cos(q1)*sin(q2)*cos(q4))
    *cos(q6)
    -((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q5)
      +((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
	-cos(q1)*sin(q2)*sin(q4))
      *cos(q5))
    *sin(q6);




  //J(2,7)
  J[13] = (-(cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	     -sin(q1)*sin(q2)*cos(q4))
    *cos(q6)
    -((cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q5)
      +((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
	-sin(q1)*sin(q2)*sin(q4))
      *cos(q5))
    *sin(q6);




  //J(3,7)
  J[20] = (sin(q2)*cos(q3)*sin(q4)-cos(q2)*cos(q4))*cos(q6)
    -(sin(q2)*sin(q3)*sin(q5)+(-cos(q2)*sin(q4)-sin(q2)*cos(q3)*cos(q4))
      *cos(q5))
    *sin(q6);




  //==========================================================


  return;

}
//======================================================================
void PCube::OrientationJacobianEndeffector(const float* const vTh, float* const J)
{


  initAngle(vTh);

  
  //====================================================================
   
  //J(1,1)
  J[0] = cos(q2)*(cos(q4)*sin(q6)*sin(q7)
		     -sin(q4)*(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7)))
    -sin(q2)*(cos(q3)*(cos(q4)
		       *(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
		       +sin(q4)*sin(q6)*sin(q7))
	      -sin(q3)*(-sin(q5)*cos(q6)*sin(q7)
			-cos(q5)*cos(q7)));


  //J(2,1)
  J[7] = cos(q2)*(-sin(q4)*(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		     -cos(q4)*sin(q6)*cos(q7))
    -sin(q2)*(cos(q3)*(cos(q4)
		       *(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		       -sin(q4)*sin(q6)*cos(q7))
	      -sin(q3)*(sin(q5)*cos(q6)*cos(q7)
			-cos(q5)*sin(q7)));


  //J(3,1)
  J[14] = cos(q2)*(cos(q4)*cos(q6)-sin(q4)*cos(q5)*sin(q6))
    -sin(q2)*(cos(q3)*(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
	      -sin(q3)*sin(q5)*sin(q6));



  //===============================

    
  //J(1,2)
  J[1] = sin(q3)*(cos(q4)*(sin(q5)*cos(q7)-cos(q5)*cos(q6)*sin(q7))
		     +sin(q4)*sin(q6)*sin(q7))
    +cos(q3)*(-sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7));


  //J(2,2)
  J[8] = sin(q3)*(cos(q4)*(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
		     -sin(q4)*sin(q6)*cos(q7))
    +cos(q3)*(sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7));



  //J(3,2)
  J[15] = sin(q3)*(cos(q4)*cos(q5)*sin(q6)+sin(q4)*cos(q6))
    +cos(q3)*sin(q5)*sin(q6);



  //===============================


  //J(1,3)
  J[2] = cos(q4)*sin(q6)*sin(q7)-sin(q4)
    *(sin(q5)*cos(q7)
      -cos(q5)*cos(q6)*sin(q7));


  //J(2,3)
  J[9] = -sin(q4)*(sin(q5)*sin(q7)+cos(q5)*cos(q6)*cos(q7))
    -cos(q4)*sin(q6)*cos(q7);


  //J(3,3)
  J[16] = cos(q4)*cos(q6)-sin(q4)*cos(q5)*sin(q6);


  //===============================

  //J(1,4)
  J[3] = -sin(q5)*cos(q6)*sin(q7)-cos(q5)*cos(q7);

  //J(2,4)
  J[10] = sin(q5)*cos(q6)*cos(q7)-cos(q5)*sin(q7);

  //J(3,4)
  J[17] = sin(q5)*sin(q6);

  //===============================

  //J(1,5)
  J[4] = sin(q6)*sin(q7);

  //J(2,5)
  J[11] = -sin(q6)*cos(q7);

  //J(3,5)
  J[18] = cos(q6);


  //===============================

  //J(1,6)
  J[5] = -cos(q7);

  //J(2,6)
  J[12] = -sin(q7);

  //J(3,6)
  J[19] = 0;


  //===============================

  //J(1,7)
  J[6] = 0;

  //J(2,7)
  J[13] = 0;

  //J(3,7)
  J[20] = -1; 

  //====================================================================
   

  return;

}
//======================================================================
void PCube::OrientationJacobianCamera(const float* const vTh, float* const J)
{


  OrientationJacobianEndeffector(vTh,m_mJoe);


  // RT * J {multiplication is done along row-wise of R matrix}
  multiply(m_mR8,m_mJoe,J,NO,NL,NO,TRANS);

#ifdef DEBUG
  std::cout<<"Orientation Jacobian End-effector\n";
  display(m_mJoe,NO,NL);
  std::cout<<"Orientation Jacobian Camera \n";
  display(J,NO,NL);
#endif


  //===============================
  return;

}
//=========================================================================
void PCube::Jacobian(const float* const vTh, float* const J)
{

  PositionJacobian(vTh, J);
  OrientationJacobian(vTh, (J+(NC*NL) ) );

#ifdef DEBUG
  std::cout<<"Jacobian\n";
  display(J,(NC+NO),NL);

#endif


  //===============================
  return ;

}

//=======================================================================
void PCube::JacobianEndeffector(const float* const vTh, float* const J)
{

  PositionJacobianEndeffector(vTh, J);
  OrientationJacobianEndeffector(vTh, (J + (NC*NL) ) );


#ifdef DEBUG
  std::cout<<"Camera Jacobian\n";
  display(J,(NC+NO),NL);

#endif


  //===============================
  return ;

}

//=======================================================================
void PCube::JacobianCamera(const float* const vTh, float* const J)
{

  PositionJacobianCamera(vTh,J);
  OrientationJacobianCamera(vTh, (J+(NC*NL)));


#ifdef DEBUG
  std::cout<<"Camera Jacobian\n";
  display(J,(NC+NO),NL);

#endif

  //===============================

  return ;

}

//=======================================================================
void PCube::JointPosition(const float* const vTh, float * const x)
{



  initAngle(vTh);


  //===============================
  //base level
  x[0] = 0;
  x[1] = 0;
  x[2] = 0;

  //===============================

  //Distal end First link
  x[3] = 0;
  x[4] = 0;
  x[5] = d1;

  //===============================

  //Distal end of Third  link
  x[6] = d3*cos(q1)*sin(q2);
  x[7] = d3*sin(q1)*sin(q2);
  x[8] = d3*cos(q2)+d1;


  //===============================

  //Distal end of Fifth link
  x[9] = d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
		+cos(q1)*sin(q2)*cos(q4))
    +d3*cos(q1)*sin(q2);


  x[10] =d5*((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	       +sin(q1)*sin(q2)*cos(q4))
    +d3*sin(q1)*sin(q2);




  x[11] = d5*(cos(q2)*cos(q4)-sin(q2)*cos(q3)*sin(q4))+d3*cos(q2)+d1;



  //===============================

  //Tip of End-effector

  x[12] = d7*(((-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q5)
		 +((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
		   -cos(q1)*sin(q2)*sin(q4))
		 *cos(q5))
		*sin(q6)
		-(-(cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
		  -cos(q1)*sin(q2)*cos(q4))
		*cos(q6))
    +d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
	 +cos(q1)*sin(q2)*cos(q4))+d3*cos(q1)*sin(q2);





  x[13] = d7*(((cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q5)
		 +((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
		   -sin(q1)*sin(q2)*sin(q4))
		 *cos(q5))
		*sin(q6)
		-(-(cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
		  -sin(q1)*sin(q2)*cos(q4))
		*cos(q6))
    +d5*((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*sin(q4)
	 +sin(q1)*sin(q2)*cos(q4))+d3*sin(q1)*sin(q2);




  x[14] = d7*((sin(q2)*sin(q3)*sin(q5)+(-cos(q2)*sin(q4)
					  -sin(q2)*cos(q3)*cos(q4))
		 *cos(q5))
		*sin(q6)
		-(sin(q2)*cos(q3)*sin(q4)-cos(q2)*cos(q4))*cos(q6))
    +d5*(cos(q2)*cos(q4)-sin(q2)*cos(q3)*sin(q4))+d3*cos(q2)+d1;


  //===============================

  return;
}

//=====================================================================
void PCube::CameraRotationMatrix(const float* const vTh, float* const RI)
{

  RotationMatrix(vTh);

  multiply(m_mR,m_mR8,RI,NO,NO,NO);

#ifdef DEBUG
  std::cout<<"End-effector Rotation Matrix \n";
  display(m_mR,NO,NO);

  std::cout<<"Camera Frame Rotation Matrix \n";
  display(m_mR8,NO,NO);


  std::cout<<"Camera Rotation Matrix \n";
  display(RI,NO,NO);

#endif

  return;
}
//=====================================================================
void PCube::CameraTransformationMatrix(const float* const vTh, float* const TI)
{

  TransformationMatrix(vTh);

  multiply(m_mT,m_mT8,TI, (NC+1),(NC+1),(NC+1));

#ifdef DEBUG
  std::cout<<"End-effector Transformation Matrix \n";
  display(m_mT,(NC+1),(NC+1));

  std::cout<<"Camera Frame Transformation Matrix \n";
  display(m_mT8,(NC+1),(NC+1));


  std::cout<<"Camera Transformation Matrix \n";
  display(TI,(NC+1),(NC+1));

#endif


  return;
}
//=====================================================================
void PCube::InverseTransformationMatrix(const float* const vTh, float* const mTI)
{

  float *vx=NULL;
  float *vxt=NULL;

  vx = allocate<float>(NC);
  vxt = allocate<float>(NC);
  //===============================
  TransformationMatrix(vTh);

  //Rotation Matrix
  for(uint i=0;i<NO;i++)
    for(uint j=0;j<NO;j++)
      mTI[i * (NO+1) + j] = m_mT[j * (NO+1) + i];

  //===============================
  //Projective portion
  mTI[3 * (NO+1) + 0] = 0;
  mTI[3 * (NO+1) + 1] = 0;
  mTI[3 * (NO+1) + 2] = 0;
  mTI[3 * (NO+1) + 3] = 1;

  //===============================
  //Inverse Position Kinematics
  PositionKinematics(vTh,vx);

  //xt = -R^T*x
  multiply(m_mR,vx,vxt,NO,1,NO,TRANS);
  negate(vxt,3);

  //Position Kinematics
  mTI[0 * (NO+1) + 3] = vxt[0];
  mTI[1 * (NO+1) + 3] = vxt[1];
  mTI[2 * (NO+1) + 3] = vxt[2];

  //===============================

  del(vx);
  del(vxt);
  return ;
}
//=====================================================================
void PCube::InverseTransformationMatrix(float* const mT)const
{


  float *vx = NULL;
  float *vxt = NULL;
  float *mR = NULL;
  float *mTI = NULL;

  vx = allocate<float>(NC);
  vxt = allocate<float>(NC);
  mR = allocate<float>(NO*NO);
  mTI = allocate<float>( (NO+1) * (NO+1) );


  //===============================

  for(uint i=0;i<NO;i++)
    for(uint j=0;j<NO;j++)
      {
	mTI[i * (NO+1) + j] = mT[j * (NO+1) + i];
	mR[i * (NO) + j] = mT[j * (NO+1) + i];
      }


  //===============================
  mTI[3 * (NO+1) + 0] = 0;
  mTI[3 * (NO+1) + 1] = 0;
  mTI[3 * (NO+1) + 2] = 0;
  mTI[3 * (NO+1) + 3] = 1;

  //===============================
  vx[0] = mT[0 * (NO+1) + 3];
  vx[1] = mT[1 * (NO+1) + 3];
  vx[2] = mT[2 * (NO+1) + 3];


  
  //xt = -R^T*x
  multiply(mR,vx,vxt,NO,1,NO,TRANS);
  negate(vxt,3);



  mTI[0 * (NO+1) + 3] = vxt[0];
  mTI[1 * (NO+1) + 3] = vxt[1];
  mTI[2 * (NO+1) + 3] = vxt[2];

  //===============================
  copy(mTI,mT,(NO+1),(NO+1));

  del(vx);
  del(vxt);
  del(mR);
  del(mTI);

  return;
}
//=====================================================================
void PCube::PositionJacobian5(const float* const vTh, float* const J)
{

  initAngle(vTh);

  //===============================
  //Jacobian of x-axis of 5th Joint
  J[0] = d5*((-cos(q1)*sin(q3)-sin(q1)*cos(q2)*cos(q3))*sin(q4)
            -sin(q1)*sin(q2)*cos(q4))
  -d3*sin(q1)*sin(q2);


  J[1] = d5*(cos(q1)*cos(q2)*cos(q4)-cos(q1)*sin(q2)*cos(q3)*sin(q4))
  +d3*cos(q1)*cos(q2);

  J[2] = d5*(-cos(q1)*cos(q2)*sin(q3)-sin(q1)*cos(q3))*sin(q4);
 
  J[3] = d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*cos(q4)
	   -cos(q1)*sin(q2)*sin(q4));


  J[4] = 0;

  J[5] = 0;

  J[6] = 0;

  //===============================
  //Jacobian of y-axis of 5th Joint

  J[7] = d5*((cos(q1)*cos(q2)*cos(q3)-sin(q1)*sin(q3))*sin(q4)
            +cos(q1)*sin(q2)*cos(q4))
    +d3*cos(q1)*sin(q2);

  J[8] = d5*(sin(q1)*cos(q2)*cos(q4)-sin(q1)*sin(q2)*cos(q3)*sin(q4))
    +d3*sin(q1)*cos(q2);

  J[9] = d5*(cos(q1)*cos(q3)-sin(q1)*cos(q2)*sin(q3))*sin(q4);

  J[10] = d5*((cos(q1)*sin(q3)+sin(q1)*cos(q2)*cos(q3))*cos(q4)
            -sin(q1)*sin(q2)*sin(q4));

  J[11] = 0;

  J[12] = 0;

  J[13] = 0;

  //===============================
  //Jacobian of z-axis of 5th Joint

  J[14] = 0;

  J[15] = d5*(-cos(q2)*cos(q3)*sin(q4)-sin(q2)*cos(q4))-d3*sin(q2);

  J[16] = d5*sin(q2)*sin(q3)*sin(q4);

  J[17] = d5*(-cos(q2)*sin(q4)-sin(q2)*cos(q3)*cos(q4));

  J[18] = 0;

  J[19] = 0;

  J[20] = 0;
  //===============================
  return ;

}
//=====================================================================
/*
void Get_Position_Velocity(const float theta[], const float thetadot[], float xdot[])
{
  int i;
  int j;
  float **Jacobian;

  Jacobian = new float* [NC];
  for(i = 0; i< NC; i++)
    Jacobian[i] = new float [NL];

  Get_Position_Jacobian(theta, Jacobian);

  for(i = 0;i<NC;i++)
    {
      xdot[i] = 0;
      for(j=0;j<NL;j++)
	xdot[i] += Jacobian[i][j] * thetadot[j];
    }


  for(uint i=0;i<NC;i++)
    delete []Jacobian[i];
  delete []Jacobian;

  return ;
}

//====================================================================
void Robot_Position_Motion(float *theta, const float *thetadot, float *x, float *xdot)
{

  float h;
  int i;
  int j;
  h = SAMPLING_TIME/SUB_INTERVAL;

  for(i=1;i<=SUB_INTERVAL;i++)
    {
      Get_Position_Velocity(theta, thetadot, xdot);

      for(j=0;j<NL;j++)
	theta[j] += thetadot[j] * h;

      for(j=0;j<NC;j++)
	x[j] += xdot[j] * h;

    }



}
//====================================================================

void Get_Orientation_Velocity(const float theta[], const float thetadot[], float xdot[])
{
  int i;
  int j;
  float **Jacobian;

  Jacobian = new float* [NO];
  for(i = 0; i< NO; i++)
    Jacobian[i] = new float [NL];

  Get_Orientation_Jacobian(theta, Jacobian);

  for(i = 0;i<NO;i++)
    {
      xdot[i] = 0;
      for(j=0;j<NL;j++)
	xdot[i] += Jacobian[i][j] * thetadot[j];
    }

  for(uint i=0;i<NO;i++)
    delete []Jacobian[i];
  delete []Jacobian;
  return ;
}

//====================================================================

void Robot_Orientation_Motion(float *theta, const float *thetadot, float *x, float *xdot)
{

  float h;
  int i;
  int j;
  h = SAMPLING_TIME/SUB_INTERVAL;

  for(i=1;i<=SUB_INTERVAL;i++)
    {
      Get_Orientation_Velocity(theta, thetadot, xdot);

      for(j=0;j<NL;j++)
	theta[j] += thetadot[j] * h;

      for(j=0;j<NO;j++)
	x[j] += xdot[j] * h;

    }



}
//====================================================================
void Get_Robot_Velocity(const float theta[], const float thetadot[], float xdot[])
{
  int i;
  int j;
  float **Jacobian;

  Jacobian = new float* [NC+NO];
  for(i = 0; i< (NC+NO); i++)
    Jacobian[i] = new float [NL];

  Get_Jacobian(theta, Jacobian);

  for(i = 0;i<(NC+NO);i++)
    {
      xdot[i] = 0;
      for(j=0;j<NL;j++)
	xdot[i] += Jacobian[i][j] * thetadot[j];
    }

  for(i = 0;i < (NC+NO);i++)
    delete []Jacobian[i];
  delete []Jacobian;
  return ;
}

//====================================================================

void Robot_Motion(float *theta, const float *thetadot, float *x, float *xdot)
{

  float h;
  int i;
  int j;
  h = SAMPLING_TIME/SUB_INTERVAL;

  for(i=1;i<=SUB_INTERVAL;i++)
    {
      Get_Robot_Velocity(theta, thetadot, xdot);

      for(j=0;j<NL;j++)
	theta[j] += thetadot[j] * h;

      for(j=0;j<(NC+NO);j++)
	x[j] += xdot[j] * h;

    }



}
//====================================================================
*/
//EOF
 