/* ----------------------------------------------------------
 * Header file for PowerCube Manipulator(with orientation)
 * available at IIT-Kanpur
 * Date: 29 December 2009

03.03.2011
The translational velocity jacobian of the 5th link in included


 * ---------------------------------------------------------- */
#ifndef _PC_
#define _PC_

#include<constants.h>


#define NL (7)                                // Number of Links
#define NC (3)                                // Number of Position Coordinates (x,y,z)
#define NO (3)                                // Number of Orientation Coordiantes (roll, pitch, yaw)


#define A8 (0)
#define ALPHA8 (0)
#define D8 (0)
#define THETA8 (PI/2)


//=====================================================================

class PCube
{

 private:

  float d1;
  float d3;
  float d5;
  float d7; 

  float q1;
  float q2;
  float q3;
  float q4;
  float q5;
  float q6;
  float q7;


  float *m_mR8;
  float *m_mT8;

  float a8;
  float d8;
  float alpha8;
  float theta8;


  float *m_mR;
  float *m_mT;


  float *m_mJp;
  float *m_mJo;

  float *m_mJpe;
  float *m_mJoe;


  float *m_vplaneNormal;
  float *m_vplanePt;


  void createBuffer();
  void initDimensionConstants();

  void PCubeDisplay();


  void initAngle(const float th[]);


 protected:



 public:


  PCube();
  ~PCube();


  void PositionKinematics(const float* const th, float* const pos);

  void RotationMatrix (const float* const vTh);
  void TransformationMatrix (const float* const vTh);


  void RotationMatrix (const float* const vTh, float* const mB);
  void TransformationMatrix (const float* const vTh, float* const mT);

  void TransformationMatrix(const float& length, const float& twist, const float& offset, const float& angle, float* constT)const;

  void RotationMatrix(const float& twist, const float& angle, float* const R)const;


  void PositionJacobian(const float* const vTh, float* const J);
  void PositionJacobianEndeffector(const float* const vTh, float* const J);
  void PositionJacobianCamera(const float* const vTh, float* const J);

  void OrientationJacobian(const float* const vTh, float* const J);
  void OrientationJacobianEndeffector(const float* const vTh, float* const J);
  void OrientationJacobianCamera(const float* const vTh, float* const J);

  void Jacobian(const float* const vTh, float* const J);
  void JacobianEndeffector(const float* const vTh, float* const J);
  void JacobianCamera(const float* const vTh, float* const J);


  void JointPosition(const float* const vTh, float* const x);

  void CameraRotationMatrix(const float* const vTh, float* const RI);
  void CameraTransformationMatrix(const float* const vTh, float* const TImage);

  void InverseTransformationMatrix(const float* const vTh, float* const mTI);
  void InverseTransformationMatrix(float* const mT)const;


  void PositionJacobian5(const float* const vTh, float* const J);



};


//=====================================================================
#endif






//=====================================================================
//EOF


/*


#define SAMPLING_TIME (0.05)                 // Sampling Interval (50ms)
#define SUB_INTERVAL  (10)                   // Sub Interval within Sampling time


  void Get_Position_Velocity(const float theta[], const float thetadot[], float xdot[]);
  void Robot_Position_Motion(float *theta, const float *thetadot, float *x, float *xdot);

  void Get_Orientation_Velocity(const float theta[], const float thetadot[], float xdot[]);
  void Robot_Orientation_Motion(float *theta, const float *thetadot, float *x, float *xdot);

  void Get_Robot_Velocity(const float theta[], const float thetadot[], float xdot[]);
  void Robot_Motion(float *theta, const float *thetadot, float *x, float *xdot);




*/
