/*
//====================================================================
File to perform matrix operations using gsl
16.01.2009
 i) Psuedo Inverse computation routine is included

//====================================================================

 */

#ifndef _PINV_
#define _PINV_


#include <cmath>
#include <new>
#include <cstdlib>
#include <iostream>


#include<gsl/gsl_matrix.h>
#include<gsl/gsl_eigen.h>
#include<gsl/gsl_linalg.h>
#include<gsl/gsl_blas.h>
#include<gsl/gsl_permutation.h>

using namespace std;
//====================================================================
void PInverse(float **J, int m, int n,  float **JI)
{
  int i;
  int j;

  gsl_matrix *A = gsl_matrix_alloc(m,n);
  gsl_matrix *AT = gsl_matrix_alloc(n,m);
  gsl_matrix *U = gsl_matrix_alloc(m,m);
  gsl_matrix *V = gsl_matrix_alloc(m,m);
  gsl_vector *S = gsl_vector_alloc(m);
  gsl_vector *W1 = gsl_vector_alloc(m);


  gsl_matrix *UT = gsl_matrix_alloc(m,m);

  gsl_vector *SI = gsl_vector_alloc(m);
  gsl_matrix *SIm = gsl_matrix_alloc(m,m);
  gsl_matrix *VSI = gsl_matrix_alloc(m,m);
  gsl_matrix *AAT = gsl_matrix_alloc(m,m);
  gsl_matrix * AATI = gsl_matrix_alloc(m,m);

  gsl_matrix *I = gsl_matrix_alloc(m,m);
  gsl_matrix *AI = gsl_matrix_alloc(n,m);


  gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc(m);
  gsl_vector_complex *eig = gsl_vector_complex_alloc(m);
  gsl_matrix_complex *evec = gsl_matrix_complex_alloc(m,m);

  //==============================================
  // Computation of AA^T
  for(i = 0;i<m;i++)
    for(j = 0;j<n;j++)
      gsl_matrix_set(A,i,j,J[i][j]);


  gsl_matrix_transpose_memcpy(AT,A);
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, A, AT,0.0, AAT);

  //=============================================
  //Computation of Inverse of AA^T
  gsl_linalg_SV_decomp(AAT,V,S,W1);

  gsl_matrix_memcpy(U,AAT);
  gsl_matrix_transpose_memcpy(UT, U);
 
      
  gsl_matrix_set_zero(SIm);
  for(i = 0;i<m;i++)
    {
      float tmp;
      tmp = gsl_vector_get(S,i);
      if( tmp > 0.0001)
	gsl_vector_set(SI,i,1/tmp);
      else
	gsl_vector_set(SI,i, 0.0);

      gsl_matrix_set(SIm, i, i, gsl_vector_get(SI,i) );
    }

   
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, SIm,0.0,  VSI);
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, VSI, UT, 0.0, AATI);
  //============================================
  // Computation of pseudo inverse
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, AT, AATI, 0.0, AI);

  for(i = 0; i < n; i++)
    for(j = 0; j< m; j++)
      JI[i][j] = gsl_matrix_get(AI,i,j);

  //=============================================
  //Result Verification 

  /*
    cout << "Pseudo-Inverse : \n";
    for(i = 0; i < n; i++)
    {
    for(j = 0; j< m; j++)
    cout<<  gsl_matrix_get(AI,i,j) << "\t";
    cout<<endl;
    }
  */

  
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, A, AI, 0.0, I);

  /*
    cout << "Identity matrix : \n";
    for(i = 0; i < m; i++)
    {
    for(j = 0; j< m; j++)
    cout<<  gsl_matrix_get(I,i,j) << "\t";
    cout<<endl;
    }
  */
  gsl_eigen_nonsymmv(I,eig,evec,w); 
  gsl_eigen_nonsymmv_sort(eig,evec,GSL_EIGEN_SORT_ABS_ASC);

  /*
   cout<<"Eigen Values:\t";
  for(i = 0;i <m;i++)
    {
      gsl_complex c  =  gsl_vector_complex_get(eig,i);
      cout<< GSL_REAL(c) << "\t";
    }
  
  cout<< endl;
  */
  //=============================================
  //Buffer Clearing
  gsl_eigen_nonsymmv_free(w);
  gsl_vector_complex_free(eig);
  gsl_matrix_complex_free(evec);


  gsl_vector_free(S);
  gsl_vector_free(SI);
  gsl_vector_free(W1);


  gsl_matrix_free(A);
  gsl_matrix_free(AT);
  gsl_matrix_free(U);
  gsl_matrix_free(V);

  gsl_matrix_free(UT);
  gsl_matrix_free(SIm);
  gsl_matrix_free(VSI);
  gsl_matrix_free(AAT);
  gsl_matrix_free(AATI);

  gsl_matrix_free(I);
  gsl_matrix_free(AI);



  return;



}
//=====================================================================
void inverse(float **A, int n, float **AI)
{

  int i;
  int j;

  int *signum;

  signum = new int [n];

  gsl_matrix *gA = gsl_matrix_alloc(n,n);
  gsl_matrix *gAI = gsl_matrix_alloc(n,n);
  gsl_matrix *I = gsl_matrix_alloc(n,n);

  gsl_permutation *p = gsl_permutation_alloc(n);


  gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc(n);
  gsl_vector_complex *eig = gsl_vector_complex_alloc(n);
  gsl_matrix_complex *evec = gsl_matrix_complex_alloc(n,n);


  for(i = 0;i<n;i++)
    for(j = 0;j<n;j++)
      gsl_matrix_set(gA,i,j,A[i][j]);




  gsl_linalg_LU_decomp(gA,p,signum);
  gsl_linalg_LU_invert(gA,p,gAI);



  for(i = 0; i < n; i++)
    for(j = 0; j< n; j++)
      AI[i][j] = gsl_matrix_get(gAI,i,j);





  //=============================================
  //Result Verification 

  /*
    cout << "Inverse: \n";
    for(i = 0; i < n; i++)
    {
    for(j = 0; j< n; j++)
    cout<<  gsl_matrix_get(gAI,i,j) << "\t";
    cout<<endl;
    }
  */
  
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, gA, gAI, 0.0, I);

  /*
    cout << "Identity matrix : \n";
    for(i = 0; i < n; i++)
    {
    for(j = 0; j< n; j++)
    cout<<  gsl_matrix_get(I,i,j) << "\t";
    cout<<endl;
    }
  */

  gsl_eigen_nonsymmv(I,eig,evec,w); 
  gsl_eigen_nonsymmv_sort(eig,evec,GSL_EIGEN_SORT_ABS_ASC);

  cout<<"Eigen Values:\t";
  for(i = 0;i <n;i++)
    {
      gsl_complex c  =  gsl_vector_complex_get(eig,i);
      cout<< GSL_REAL(c) << "\t";
    }
  
  cout<< endl;

  //=============================================

  gsl_eigen_nonsymmv_free(w);
  gsl_vector_complex_free(eig);
  gsl_matrix_complex_free(evec);


  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_matrix_free(I);
  gsl_permutation_free(p);

  delete []signum;

}
//=====================================================================
void right_inverse(float **A, int m, int n, float **AI)
{

  int i;
  int j;

  float **AAT;
  float **AATI;

  AAT = new float *[m];
  for(int i = 0;i<m;i++)
    AAT[i] = new float [m];


  AATI = new float *[m];
  for(int i = 0;i<m;i++)
    AATI[i] = new float [m];

  gsl_matrix *gA    = gsl_matrix_alloc(m,n);
  gsl_matrix *gAI   = gsl_matrix_alloc(n,m);
  gsl_matrix *gAAT  = gsl_matrix_alloc(m,m);
  gsl_matrix *gAATI = gsl_matrix_alloc(m,m);


  for(i = 0;i<m;i++)
    for(j = 0;j<n;j++)
      gsl_matrix_set(gA,i,j,A[i][j]);

  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, gA, gA,0.0,  gAAT);

  for(i = 0; i < m; i++)
    for(j = 0; j< m; j++)
      AAT[i][j] = gsl_matrix_get(gAAT,i,j);

  inverse(AAT,m,AATI);

  for(i = 0;i<m;i++)
    for(j = 0;j<m;j++)
      gsl_matrix_set(gAATI,i,j,AATI[i][j]);

  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, gA, gAATI,0.0,  gAI);


  for(i = 0; i < n; i++)
    for(j = 0; j< m; j++)
      AI[i][j] = gsl_matrix_get(gAI,i,j);



  delete []AAT;
  delete []AATI;

  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_matrix_free(gAAT);
  gsl_matrix_free(gAATI);


}
//=====================================================================
void left_inverse(float **A, int m, int n, float **AI)
{

  int i;
  int j;

  float **ATA;
  float **ATAI;

  ATA = new float *[n];
  for(int i = 0;i<n;i++)
    ATA[i] = new float [n];


  ATAI = new float *[n];
  for(int i = 0;i<n;i++)
    ATAI[i] = new float [n];

  gsl_matrix *gA    = gsl_matrix_alloc(m,n);
  gsl_matrix *gAI   = gsl_matrix_alloc(n,m);
  gsl_matrix *gATA  = gsl_matrix_alloc(n,n);
  gsl_matrix *gATAI = gsl_matrix_alloc(n,n);


  for(i = 0;i<m;i++)
    for(j = 0;j<n;j++)
      gsl_matrix_set(gA,i,j,A[i][j]);

  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, gA, gA,0.0,  gATA);

  for(i = 0; i < n; i++)
    for(j = 0; j< n; j++)
      ATA[i][j] = gsl_matrix_get(gATA,i,j);

  inverse(ATA,n,ATAI);

  for(i = 0;i<n;i++)
    for(j = 0;j<n;j++)
      gsl_matrix_set(gATAI,i,j,ATAI[i][j]);

  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, gATAI, gA,0.0,  gAI);


  for(i = 0; i < n; i++)
    for(j = 0; j< m; j++)
      AI[i][j] = gsl_matrix_get(gAI,i,j);



  delete []ATA;
  delete []ATAI;

  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_matrix_free(gATA);
  gsl_matrix_free(gATAI);


}
//=====================================================================

#endif
