#include<operations.h>


//====================================================================
void copy(const float* const& A, gsl_matrix* const gA, const uint& row, const uint& col, const uint& ATrans /*= NO_TRANS*/)
{


  if(ATrans == NO_TRANS)
    {
      for(uint i = 0;i<row;i++)
	for(uint j = 0;j<col;j++)
	  gsl_matrix_set(gA,i,j,A[i*col+j]);
    }
  else
    {

      for(uint i = 0;i<row;i++)
	for(uint j = 0;j<col;j++)
	  gsl_matrix_set(gA,j,i,A[i*col+j]);

    }

  //================================

  return;
}
//====================================================================
void copy(const gsl_matrix* const& gA, float* const& A, const uint& row, const uint& col, const uint& ATrans /*= NO_TRANS*/)
{



  if(ATrans == NO_TRANS)
    {
      for(uint i = 0; i < row; i++)
	for(uint j = 0; j< col; j++)
	  A[i*col+j] = gsl_matrix_get(gA,i,j);
    }
  else
    {

      for(uint i = 0; i < row; i++)
	for(uint j = 0; j< col; j++)
	  A[i*col+j] = gsl_matrix_get(gA,j,i);

    }

  //================================


  return;
}
//====================================================================
void copy(const float* const& A, gsl_vector* const gA, const uint& col)
{

  for(uint j = 0;j<col;j++)
    gsl_vector_set(gA,j,A[j]);

  //================================

  return;
}
//====================================================================
void copy(const gsl_vector* const& gA, float* const& A, const uint& col)
{


  for(uint j = 0; j< col; j++)
    A[j] = gsl_vector_get(gA,j);

  //================================

  return;
}
//====================================================================
void display(const gsl_matrix* const& gA, const uint& row, const uint& col)
{

  for(uint i = 0; i < row; i++)
    {
      for(uint j = 0; j< col; j++)
	std::cout<<  gsl_matrix_get(gA,i,j) << "\t";
      newline();
    }

  //================================

  return;
}
//====================================================================
void display(const gsl_vector* const& gA,const uint& col)
{

  for(uint i = 0; i < col; i++)
    {
      std::cout<<  gsl_vector_get(gA,i) << "\t";
    }

  newline();

  //================================

  return;
}
//====================================================================
void PInverse(const float* const J, const uint& row, const uint& col,  float* const JI)
{
  uint i=0;
  uint j=0;

  gsl_matrix *A = gsl_matrix_alloc(row,col);
  gsl_matrix *AT = gsl_matrix_alloc(col,row);
  gsl_matrix *U = gsl_matrix_alloc(row,row);
  gsl_matrix *V = gsl_matrix_alloc(row,row);
  gsl_vector *S = gsl_vector_alloc(row);
  gsl_vector *W1 = gsl_vector_alloc(row);


  gsl_matrix *UT = gsl_matrix_alloc(row,row);

  gsl_vector *SI = gsl_vector_alloc(row);
  gsl_matrix *SIm = gsl_matrix_alloc(row,row);
  gsl_matrix *VSI = gsl_matrix_alloc(row,row);
  gsl_matrix *AAT = gsl_matrix_alloc(row,row);
  gsl_matrix * AATI = gsl_matrix_alloc(row,row);

  gsl_matrix *AI = gsl_matrix_alloc(col,row);


  //================================
  // Computation of AA^T

  copy(J,A,row,col);

  gsl_matrix_transpose_memcpy(AT,A);
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, A, AT,0.0, AAT);

  //================================
  //Computation of Inverse of AA^T
  gsl_linalg_SV_decomp(AAT,V,S,W1);

  gsl_matrix_memcpy(U,AAT);
  gsl_matrix_transpose_memcpy(UT, U);
 
  //================================
      
  gsl_matrix_set_zero(SIm);
  for(i = 0;i<row;i++)
    {
      float tmp;
      tmp = gsl_vector_get(S,i);
      if( tmp > 0.00001)
	gsl_vector_set(SI,i,1/tmp);
      else
	gsl_vector_set(SI,i, 0.0);

      gsl_matrix_set(SIm, i, i, gsl_vector_get(SI,i) );
    }

   
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, SIm,0.0,  VSI);
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, VSI, UT, 0.0, AATI);
  //================================
  // Computation of pseudo inverse
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, AT, AATI, 0.0, AI);


  copy(AI, JI, col,row);

  //================================
  //Result Verification 

#ifdef DEBUG

  {
    gsl_matrix *I = gsl_matrix_alloc(row,row);
    gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc(row);

    gsl_vector_complex *eig = gsl_vector_complex_alloc(row);
    gsl_matrix_complex *evec = gsl_matrix_complex_alloc(row,row);


    //=============================
    copy(J,A,row,col);

    //===============================


  
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, A, AI, 0.0, I);
 

    gsl_eigen_nonsymmv(I,eig,evec,w); 
    gsl_eigen_nonsymmv_sort(eig,evec,GSL_EIGEN_SORT_ABS_ASC);

    //===============================

    std::cout << "Pseudo-Inverse : \n";
    display(AI,col,row);

    std::cout << "Identity matrix : \n";
    display(I,row,row);


    std::cout<<"Eigen Values:\t";
    for(i = 0;i <row;i++)
      {
	gsl_complex c  =  gsl_vector_complex_get(eig,i);
	std::cout<< GSL_REAL(c) << "\t";
      }
  
    newline();

    //===============================


    //Verification Buffers clearing
    gsl_eigen_nonsymmv_free(w);
    gsl_vector_complex_free(eig);
    gsl_matrix_complex_free(evec);
    gsl_matrix_free(I);
  }

#endif
  //================================
  //Buffer Clearing


  gsl_vector_free(S);
  gsl_vector_free(SI);
  gsl_vector_free(W1);


  gsl_matrix_free(A);
  gsl_matrix_free(AT);
  gsl_matrix_free(U);
  gsl_matrix_free(V);

  gsl_matrix_free(UT);
  gsl_matrix_free(SIm);
  gsl_matrix_free(VSI);
  gsl_matrix_free(AAT);
  gsl_matrix_free(AATI);

  gsl_matrix_free(AI);



  return;



}
//=====================================================================
//computation of inverse with singular value decomposition
void inverseSV(const float* A, const uint& order, float* const AI)
{

  uint i=0;
  uint j=0;

  float tmp = 0.0;


  gsl_matrix *gA = gsl_matrix_alloc(order,order);
  gsl_matrix *gU = gsl_matrix_alloc(order,order);
  gsl_matrix *gV = gsl_matrix_alloc(order,order);
  gsl_vector *gS = gsl_vector_alloc(order);
  gsl_vector *gW1 = gsl_vector_alloc(order);


  gsl_matrix *gSIm = gsl_matrix_alloc(order,order);
  gsl_matrix *gVSI = gsl_matrix_alloc(order,order);

  gsl_matrix *gAI = gsl_matrix_alloc(order,order);



  //================================
  // Copying the matrix

  copy(A,gA,order,order);


  //================================
  //Computation of Inverse of AA^T
  gsl_linalg_SV_decomp(gA,gV,gS,gW1);

  gsl_matrix_memcpy(gU,gA);
 
    //===============================
      
  gsl_matrix_set_zero(gSIm);
  for(i = 0;i<order;i++)
    {
      for(j=0;j<order;j++)
	if(i==j)
	  {
	    tmp = gsl_vector_get(gS,i);
	    if( tmp > 0.00001)
	      gsl_matrix_set(gSIm, i, i, 1/tmp );
	    else
	      gsl_matrix_set(gSIm, i, i, 0.0 );
	  }
	else
	  gsl_matrix_set(gSIm,i,j,0.0);
    }

   
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, gV, gSIm,0.0,  gVSI);
  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, gVSI, gU, 0.0, gAI);
  //================================
  // Computation of pseudo inverse

  copy(gAI,AI,order,order);
  //================================
  //Result Verification 

#ifdef DEBUG


  gsl_matrix *gI = gsl_matrix_alloc(order,order);
  gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc(order);

  gsl_vector_complex *eig = gsl_vector_complex_alloc(order);
  gsl_matrix_complex *evec = gsl_matrix_complex_alloc(order,order);

  //================================
  // Copying the matrix
  copy(A,gA,order,order);


  //================================

  
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, gA, gAI, 0.0, gI);


  gsl_eigen_nonsymmv(gI,eig,evec,w); 
  gsl_eigen_nonsymmv_sort(eig,evec,GSL_EIGEN_SORT_ABS_ASC);


  //================================

  std::cout << "Inverse using singular value decomposition:: \n";
  display(gAI,order,order);

  std::cout << "Identity matrix : \n";
  display(gI,order,order);


  std::cout<<"Eigen Values:\t";
  for(i = 0;i <order;i++)
    {
      gsl_complex c  =  gsl_vector_complex_get(eig,i);
      std::cout<< GSL_REAL(c) << "\t";
    }
  
  newline();

  //================================


  //Verification Buffers clearing
  gsl_eigen_nonsymmv_free(w);
  gsl_vector_complex_free(eig);
  gsl_matrix_complex_free(evec);
  gsl_matrix_free(gI);



#endif
  //================================
  //Buffer Clearing


  gsl_vector_free(gS);
  gsl_vector_free(gW1);


  gsl_matrix_free(gA);
  gsl_matrix_free(gU);
  gsl_matrix_free(gV);

  gsl_matrix_free(gSIm);
  gsl_matrix_free(gVSI);

  gsl_matrix_free(gAI);



  return;


}
//=====================================================================
//inverse with LU Decomposition
void inverseLU(const float *A, const uint& order, float* const AI)
{

  uint i = 0;
  uint j = 0;
  int *signum = NULL;

  signum = new int [order];

  gsl_matrix *gA = gsl_matrix_alloc(order,order);
  gsl_matrix *gAI = gsl_matrix_alloc(order,order);

  gsl_permutation *p = gsl_permutation_alloc(order);


  //================================


  copy(A,gA,order,order);


  gsl_linalg_LU_decomp(gA,p,signum);
  gsl_linalg_LU_invert(gA,p,gAI);


  copy(gAI,AI,order,order);


  //================================
  //Result Verification 
#ifdef DEBUG

  gsl_matrix *I = gsl_matrix_alloc(order,order);
  gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc(order);
  gsl_vector_complex *eig = gsl_vector_complex_alloc(order);
  gsl_matrix_complex *evec = gsl_matrix_complex_alloc(order,order);

  //================================
  // Copying the matrix

  copy(A,gA,order,order);
  //================================

  
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, gA, gAI, 0.0, I);


  gsl_eigen_nonsymmv(I,eig,evec,w); 
  gsl_eigen_nonsymmv_sort(eig,evec,GSL_EIGEN_SORT_ABS_ASC);

  //================================


  std::cout << "Inverse with LU Decomposition: \n";
  display(gAI,order,order);

  std::cout << "Identity matrix : \n";
  display(I,order,order);

 
  std::cout<<"Eigen Values:\t";
  for(i = 0;i <order;i++)
    {
      gsl_complex c  =  gsl_vector_complex_get(eig,i);
      std::cout<< GSL_REAL(c) << "\t";
    }
  
  std::cout<< std::endl;

  //================================

  gsl_eigen_nonsymmv_free(w);
  gsl_vector_complex_free(eig);
  gsl_matrix_complex_free(evec);
  gsl_matrix_free(I);


#endif
  //================================



  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_permutation_free(p);

  delete []signum;

}
//=====================================================================
void RInverse(const float* const A, const uint& row, const uint& col, float* const AI)
{

  uint i=0;
  uint j=0;

  float *AAT = NULL;
  float *AATI = NULL;

  AAT = new float [row * row];
  AATI = new float [row * row];


  gsl_matrix *gA    = gsl_matrix_alloc(row, col);
  gsl_matrix *gAI   = gsl_matrix_alloc(col, row);
  gsl_matrix *gAAT  = gsl_matrix_alloc(row, row);
  gsl_matrix *gAATI = gsl_matrix_alloc(row, row);

  //================================


  copy(A,gA,row,col);

  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, gA, gA,0.0,  gAAT);

  copy(gAAT,AAT,row,row);

  inverseSV(AAT,row,AATI);

  copy(AATI,gAATI,row,row);

  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, gA, gAATI,0.0,  gAI);


  copy(gAI,AI,col,row);

  //================================


  delete []AAT;
  delete []AATI;

  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_matrix_free(gAAT);
  gsl_matrix_free(gAATI);

  //================================
  return;

}
//=====================================================================
void LInverse(const float* const A, const uint& row, const uint& col, float* const AI)
{

  uint i = 0;
  uint j = 0;

  float *ATA;
  float *ATAI;

  ATA = new float [col * col];
  ATAI = new float [col * col];

  gsl_matrix *gA    = gsl_matrix_alloc(row, col);
  gsl_matrix *gAI   = gsl_matrix_alloc(col, row);
  gsl_matrix *gATA  = gsl_matrix_alloc(col, col);
  gsl_matrix *gATAI = gsl_matrix_alloc(col, col);

  //================================


  copy(A,gA,row,col);

  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, gA, gA,0.0,  gATA);


  copy(gATA,ATA,col,col);

  inverseSV(ATA,col,ATAI);

  copy(ATAI,gATAI,col,col);

  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, gATAI, gA,0.0,  gAI);


  copy(gAI,AI,col,row);

  //================================

  delete []ATA;
  delete []ATAI;

  gsl_matrix_free(gA);
  gsl_matrix_free(gAI);
  gsl_matrix_free(gATA);
  gsl_matrix_free(gATAI);

  //================================

  return;
}
//=====================================================================
void computeEigen(const float* const& A, const uint& order, float* const& eigenval, float* const& eigvec)
{

  gsl_matrix *gI = gsl_matrix_alloc(order,order);
  gsl_eigen_nonsymmv_workspace *gw = gsl_eigen_nonsymmv_alloc(order);
  gsl_vector_complex *geig = gsl_vector_complex_alloc(order);
  gsl_matrix_complex *gevec = gsl_matrix_complex_alloc(order,order);


  //================================


  copy(A,gI,order,order);


  gsl_eigen_nonsymmv(gI,geig,gevec,gw); 
  gsl_eigen_nonsymmv_sort(geig,gevec,GSL_EIGEN_SORT_ABS_ASC);


  //================================

#ifdef DEBUG

  std::cout<<"Eigen Values:\t";
  for(uint i = 0;i <order;i++)
    {
      gsl_complex c  =  gsl_vector_complex_get(geig,i);
      std::cout<< GSL_REAL(c) << "\t";
    }

  newline();  

#endif


  //===============================

  if(eigenval !=NULL)
    {

      for(uint i=0;i<order;i++)
	{
	  eigenval[i] = GSL_REAL(gsl_vector_complex_get(geig,i));
	}


    }
 

  //===============================


  if(eigvec !=NULL)
    {

      for(uint i=0;i<order;i++)
	{

	  for(uint j=0;j<order;j++)
	    {
	      eigvec[i * order + j] = GSL_REAL(gsl_matrix_complex_get(gevec,i,j));

	    }
	}

    }


  //===============================
  //Clearing the buffers


  gsl_matrix_free(gI);

  gsl_eigen_nonsymmv_free(gw);
  gsl_vector_complex_free(geig);
  gsl_matrix_complex_free(gevec);

  //================================

  return;

}
//=====================================================================
//EOF
